//
// Implementation of Echofon model objects
//
// Copyright (c) 2010 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var EXPORTED_SYMBOLS = ["User", "Status", "DirectMessage", "Thread", "List", "SavedSearch", "Blocking", "NoRetweet", "DBManager"];

var TWEET_TYPES = {
  "home":     0,
  "mentions": 1,
  "messages": 2,
  "lists":    3,
  "search":   4
};

const SCHEMA_VERSION = 1;
const MAX_STATUSES = 10000;

var JSON = Components.classes["@mozilla.org/dom/json;1"].createInstance(Components.interfaces.nsIJSON);

Components.utils.import("resource://echofon/EchofonUtils.jsm");

//
// Convert *id_str to *id (snowflake)
//
function convertIdStr(obj)
{
  for (var key in obj) {
    if (typeof obj[key] == 'object') {
      convertIdStr(obj[key]);
    }
    else if (key.match(/id$/)) {
      if (obj[key + '_str']) {
        obj[key] = obj[key + '_str'];
      }
      else {
        obj[key] = obj[key].toString();
      }
    }
  }
}

var DBManager = {
  _db: {},

  db: function(recipient_id) {

    if (!recipient_id) {
      EchofonUtils.dumpStackTrace();
      dump("DBManager:db - Undefined recipient_id\n");
      return null;
    }
    if (!this._db[recipient_id]) {
      var db = Components.classes["@echofon.com/echofon-database;1"].createInstance(Components.interfaces.nsIEchofonDatabase);
      db.openDatabase(recipient_id);

      // check schema version
      var stmt = db.prepare("PRAGMA user_version");
      stmt.step();
      var version = stmt.getInt32(0);
      stmt.finalize();
      if (version != SCHEMA_VERSION) {
        dump("DB version: " + version + " != " + SCHEMA_VERSION + "\n");
        db = this.updateSchema(recipient_id, db);
      }

      try {
        this.optimize(db);
      }
      catch (e) {
        // If optimization failed, clear database on the next launch
        Components.utils.reportError("Failed to optimize database: " + e.message);
        EchofonUtils.pref().setBoolPref("clearDB", true);
      }

      this._db[recipient_id] = db;
    }
    return this._db[recipient_id];
  },

  optimize: function(db) {
    var stmt = db.prepare("SELECT count (*) FROM statuses");
    stmt.step();
    var count = stmt.getInt32(0);
    stmt.finalize();
    if (count > MAX_STATUSES) {
      EchofonUtils.log("# of statuses is " + count + ": optimize database...");
      stmt = db.prepare("SELECT id FROM statuses ORDER BY id LIMIT 1 OFFSET " + MAX_STATUSES);
      stmt.step();
      var id = stmt.getString(0);
      stmt.finalize();

      db.exec("BEGIN;" +
              "DELETE FROM statuses WHERE id < " + id + ";" +
              "DELETE FROM home     WHERE id < " + id + ";" +
              "DELETE FROM mentions WHERE id < " + id + ";" +
              "DELETE FROM tweets   WHERE id < " + id + ";" +
              "COMMIT;VACUUM;REINDEX;");
    }
  },

  updateSchema: function(recipient_id, db) {
    db.close();

    var file = Components.classes["@mozilla.org/file/directory_service;1"].getService(Components.interfaces.nsIProperties).get("ProfD", Components.interfaces.nsIFile);
    file.append('echofon_' + recipient_id + '.sqlite');
    if (file.exists()) {
      file.remove(false);
    }
    db.openDatabase(recipient_id);
    return db;
  },

  _64bitadd: function(a, b) {
    for (var i in this._db) {
      var db = this._db[i];
      var stmt = db.prepare("SELECT ?1 + ?2");
      stmt.bindStringParameter(0, a);
      stmt.bindStringParameter(1, b);
      stmt.step();
      return stmt.getString(0);
    }
    return "0";
  },

  _64bitsub: function(a, b) {
    for (var i in this._db) {
      var db = this._db[i];
      var stmt = db.prepare("SELECT ?1 - ?2");
      stmt.bindStringParameter(0, a);
      stmt.bindStringParameter(1, b);
      stmt.step();
      return stmt.getString(0);
    }
    return "0";
  },

  calcSnowflakeId: function(id, minutes) {
    for (var i in this._db) {
      var db = this._db[i];
      var stmt = db.prepare("SELECT ((?1 >> 22) - (?2 * 60 * 1000)) << 22");
      stmt.bindStringParameter(0, id);
      stmt.bindStringParameter(1, minutes);
      stmt.step();
      return stmt.getString(0);
    }
    return id;
  },

  closeAll: function() {
    for (var i in this._db) {
      var db = this._db[i];
      db.close();
    }
    this._db = {};
  }
}

/////////////////////////////////////////////////////////////////////////
//
// User model
//
function User(json, recipient_id, stmt)
{
  this.recipient_id = recipient_id;
  if (recipient_id) {
    this.db = DBManager.db(recipient_id);
  }

  // Init from JSON
  if (json) {
    convertIdStr(json);
    for (var i in json) {
      this[i] = json[i];
    }
  }

  // Init from DB
  if (stmt) {
    this.id                = stmt.getString(0);
    this.name              = stmt.getString(1);
    this.screen_name       = stmt.getString(2);
    this.location          = stmt.getString(3);
    this.description       = stmt.getString(4);
    this.url               = stmt.getString(5);
    this.followers_count   = stmt.getInt32(6);
    this.friends_count     = stmt.getInt32(7);
    this.favorites_count   = stmt.getInt32(8);
    this.statuses_count    = stmt.getInt32(9);
    this.listed_count      = stmt.getInt32(10);
    this.profile_image_url = stmt.getString(11);
    this.protected         = stmt.getInt32(12);
    this.verified          = stmt.getInt32(13);
    this.geo_enabled       = stmt.getInt32(14);
    this.updated_at        = stmt.getInt64(15);
  }
}

User.prototype = {
  insertIntoDB: function() {
    var stmt = this.db.prepare("REPLACE INTO users VALUES (?,?,?,?,?, ?,?,?,?,?, ?,?,?,?,?, ?)");

    stmt.bindStringParameter( 0, this.id);
    stmt.bindStringParameter( 1, this.name);
    stmt.bindStringParameter( 2, this.screen_name);
    stmt.bindStringParameter( 3, this.location);
    stmt.bindStringParameter( 4, this.description);
    stmt.bindStringParameter( 5, this.url);
    stmt.bindInt32Parameter(  6, this.followers_count);
    stmt.bindInt32Parameter(  7, this.friends_count);
    stmt.bindInt32Parameter(  8, this.favorites_count);
    stmt.bindInt32Parameter(  9, this.statuses_count);
    stmt.bindInt32Parameter( 10, this.listed_count);
    stmt.bindStringParameter(11, this.profile_image_url);
    stmt.bindInt32Parameter( 12, this.protected);
    stmt.bindInt32Parameter( 13, this.verified);
    stmt.bindInt32Parameter( 14, this.geo_enabled);
    stmt.bindInt64Parameter( 15, this.updated_at);
    try {
      stmt.execute();
    }
    catch (e) {
      this.db.logError("Error on insert user");
    }
  },

  profileImage72: function() {
    return this.profile_image_url.replace(/(.+)_normal\.(png|gif|jpg|jpeg)$/, "$1_bigger.$2");
  },

  profileImageUrl128: function() {
    return this.profile_image_url.replace(/(.+)_normal\.(png|gif|jpg|jpeg)$/, "$1_reasonably_small.$2");
  },

  profileImageUrlFull: function() {
    return this.profile_image_url.replace(/(.+)_normal\.(png|gif|jpg|jpeg)$/, "$1.$2");
  },

  isFollowing: function() {
    var stmt = this.db.prepare("SELECT id FROM following WHERE id = ?");
    stmt.bindStringParameter(0, this.id);
    return stmt.executeStep() ? true : false;
  },

  toString: function() {
    return "User: " + this.screen_name + " - " + this.id;
  }
}

User.findById = function(id, recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("SELECT * FROM users WHERE user_id = ?");
  stmt.bindInt64Parameter( 0, id);

  if (stmt.executeStep()) {
    return new User(null, recipient_id, stmt);
  }
  else {
    return null;
  }
}

User.isFollowing = function(user_id, dbuid) {
  var stmt = DBManager.db(dbuid).prepare("SELECT id FROM following WHERE id = ?");
  stmt.bindStringParameter(0, user_id);
  return stmt.executeStep() ? true : false;
}

User.findByScreenName = function(name, recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("SELECT * FROM users WHERE screen_name = ?");
  stmt.bindStringParameter( 0, name);

  if (stmt.executeStep()) {
    return new User(null, recipient_id, stmt);
  }
  else {
    return null;
  }
}

User.follow = function(user_id, dbuid) {
  var db = DBManager.db(dbuid);
  var stmt = db.prepare("REPLACE INTO following VALUES (?)");
  stmt.bindInt32Parameter(0, user_id);
  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.rollback();
  }
}

User.unfollow = function(user_id, dbuid) {
  var db = DBManager.db(dbuid);
  var stmt = db.prepare("DELETE FROM following WHERE id = ?");
  stmt.bindInt32Parameter(0, user_id);
  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch(e) {
    db.rollback();
  }
}

/////////////////////////////////////////////////////////////////////////
//
// Status model
//
function Status(json, type, recipient_id)
{
  convertIdStr(json);
  for (var i in json) {
    this[i] = json[i];
  }

  // Handle retweet
  if (json && json['retweeted_status']) {
    this.retweeted_status_id   = json.retweeted_status.id;
    this.retweeter_screen_name = json.user.screen_name;
    this.retweeter_user_id     = json.user.id;
    this.retweeted_at          = json.created_at;
    var keys = ['user', 'text', 'created_at', 'source', 'geo', 'in_reply_to_status_id', 'in_reply_to_user_id', 'in_reply_to_screen_name', 'entities'];
    for (var i in keys) {
      var key = keys[i];
      if (json['retweeted_status'][key]) {
        this[key] = json['retweeted_status'][key];
      }
    }
  }

  if (json) {
    if (type == 'search' && json.from_user) {
      this.user = new User(null, recipient_id);
      this.user.screen_name = json.from_user;
      this.user.id = json.from_user_id;
      this.user.profile_image_url = json.profile_image_url;
    }
    else {
      this.user = new User(this.user, recipient_id);
    }
  }

  this.type = type;
  this.has_mention = false;
  this.recipient_id = recipient_id;
  this.db = DBManager.db(recipient_id);

  this.updateAttributes();
}

Status.prototype = {
  insertIntoDB: function() {
    if (this.type == 'lists' || this.type == 'search') return true;

    // Check duplicated retweet
    if (this.retweeted_status_id) {
      var rtcheck = this.db.prepare("SELECT id FROM statuses WHERE id = ?1 OR retweeted_status_id = ?1");
      rtcheck.bindStringParameter(0, this.retweeted_status_id);
      if (rtcheck.executeStep()) {
        return false;
      }
    }

    var stmt = this.db.prepare("INSERT INTO statuses VALUES (?,?,?,?,?, ?,?,?,?,?, ?,?,?,?,?, ?,?,?)");

    stmt.bindStringParameter(0, this.id);
    stmt.bindStringParameter( 1, this.user.id);
    stmt.bindStringParameter(2, this.text);
    stmt.bindInt64Parameter( 3, new Date(this.created_at).getTime());
    stmt.bindStringParameter(4, this.source);

    if (this.geo) {
      this.latitude  = this.geo.coordinates[0];
      this.longitude = this.geo.coordinates[1];
      stmt.bindDoubleParameter(5, this.latitude);
      stmt.bindDoubleParameter(6, this.longitude);
    }

    stmt.bindStringParameter(7, this.in_reply_to_status_id);
    stmt.bindStringParameter(8, this.in_reply_to_screen_name);

    if (this.retweeted_status) {
      stmt.bindStringParameter( 9, this.retweeted_status_id);
      stmt.bindStringParameter(10, this.retweeter_user_id);
      stmt.bindStringParameter(11, this.retweeter_screen_name);
      stmt.bindInt32Parameter( 12, new Date(this.retweeted_at).getTime());
    }
    stmt.bindInt32Parameter(13, this.favorited);
    if (this.place) {
      stmt.bindStringParameter(14, JSON.encode(this.place));
    }
    if (this.contributors) {
      stmt.bindStringParameter(15, JSON.encode(this.contributors));
    }
    if (this.entities) {
      stmt.bindStringParameter(16, JSON.encode(this.entities));
    }

    var ret = true;
    try {
      stmt.execute();
    }
    catch (e) {
      if (this.db.lastError() != 19) {
        this.db.logError("Error on insert status:" + this.id);
        return false;
      }
    }

    if (this.type) {
      stmt = this.db.prepare("INSERT INTO " + this.type + " VALUES (?)");
      stmt.bindStringParameter(0, this.id);
      try {
        stmt.execute();
      }
      catch (e) {
        if (this.db.lastError() != 19) {
          this.db.logError("Error on insert " + this.type);
          return false;
        }
        else {
          //
          // Return false if the status is already in the database (This happens when user receives mentions from followers)
          //
          ret = false;
        }
      }
    }

  if (0) {
    var fts = this.db.prepare("INSERT INTO tweets VALUES (?, ?)");
    if (fts) {
      fts.bindStringParameter(0, this.id);
      fts.bindStringParameter(1, this.text);
      try {
        fts.execute();
      }
      catch (e) {
        EchofonUtils.debug("Failed to insert tweets into FTS table");
      }
    }
  }

    this.user.insertIntoDB();
    return ret;
  },

  retweet: function(tweet) {
    this.retweeted_at = tweet.created_at;
    this.retweeter_screen_name = tweet.user.screen_name;
    this.retweeted_status_id = tweet.id_str;
    this.retweeter_user_id = tweet.user.id;

    var stmt = this.db.prepare("UPDATE statuses SET retweeted_at = ?, retweeter_screen_name = ?, retweeter_user_id = ?, retweeted_status_id = ? WHERE id = ?");

    stmt.bindInt32Parameter( 0, this.retweeted_at);
    stmt.bindStringParameter(1, this.retweeter_screen_name);
    stmt.bindInt32Parameter( 2, this.retweeter_user_id);
    stmt.bindStringParameter(3, this.retweeted_status_id);
    stmt.bindStringParameter(4, this.id);
    try {
      stmt.execute();
    }
    catch (e) {
      return false;
    }
    return true;
  },

  undoRetweet: function() {
    this.retweeted_at = 0;
    this.retweeter_screen_name = "";
    this.retweeted_status_id = "";
    this.retweeter_user_id = 0;

    var stmt = this.db.prepare("UPDATE statuses SET retweeted_at = 0, retweeter_screen_name = '', retweeter_user_id = 0, retweeted_status_id = 0 WHERE id = ?");
    stmt.bindStringParameter(0, this.id);
    try {
      stmt.execute();
    }
    catch (e) {
      return false;
    }
    return true;
  },

  updateAttributes: function() {
    // update reply, highlight, and the other attribute of this status
    //
    if (this.entities && this.entities.user_mentions) {
      for (var i in this.entities.user_mentions) {
        var m = this.entities.user_mentions[i];
        if (m.id == this.recipient_id) {
          this.has_mention = true;
          break;
        }
      }
    }
  }
};

Status.initWithStatement = function(stmt, type, recipient_id) {
  var status = new Status(null, type, recipient_id);

  // from statuses table
  status.id                      = stmt.getString(0);
  status.user_id                 = stmt.getString(1);
  status.text                    = stmt.getString(2);
  status.created_at              = stmt.getInt64(3);
  status.source                  = stmt.getString(4);

  status.latitude                = stmt.getDouble(5);
  status.longitude               = stmt.getDouble(6);

  status.in_reply_to_status_id   = stmt.getString(7);
  status.in_reply_to_screen_name = stmt.getString(8);

  status.retweeted_status_id     = stmt.getString(9);
  status.retweeter_user_id       = stmt.getString(10);
  status.retweeter_screen_name   = stmt.getString(11);
  status.retweeted_at            = stmt.getInt32(12);
  status.favorited               = stmt.getInt32(13) ? true : false;

  try { status.place             = JSON.decode(stmt.getString(14));  } catch (e) {}
  try { status.contributors      = JSON.decode(stmt.getString(15));  } catch (e) {}
  try { status.entities          = JSON.decode(stmt.getString(16));  } catch (e) {}

  status.user = User.findById(status.user_id, recipient_id);

  status.updateAttributes();

  return status;
}

Status.restore = function(recipient_id, type, count, max_id) {
  var results = [];

  if (max_id) {
    var sql = "SELECT statuses.* FROM statuses," + type + " WHERE statuses.id = " + type + ".id AND statuses.id <= ?2 ORDER BY statuses.id DESC LIMIT ?1";
  }
  else {
    var sql = "SELECT statuses.* FROM statuses," + type + " WHERE statuses.id = " + type + ".id ORDER BY statuses.id DESC LIMIT ?1";
  }

  var stmt = DBManager.db(recipient_id).prepare(sql);

  if (!count) count = 20;
  stmt.bindInt32Parameter(0, count);
  if (max_id) {
    stmt.bindStringParameter(1, max_id);
  }

  try {
    while (stmt.executeStep()) {
      results.push(Status.initWithStatement(stmt, type, recipient_id));
    }
  }
  catch (e) {
    dump(e.message);
    DBManager.db(recipient_id).logError("Failed to restore " + type + " from local cache");
  }
  stmt.finalize();

  return results;
};

Status.findById = function(status_id, recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("SELECT * FROM statuses WHERE statuses.id = ?1");
  stmt.bindStringParameter(0, status_id);
  var status = null;
  try {
    if (stmt.executeStep()) {
      status = Status.initWithStatement(stmt, 0, recipient_id);
    }
  }
  catch (e) {
    this.db.logError("Failed to get tweet: " + e.message);
  }

  return status;
};

Status.toggleFavorite = function(recipient_id, status_id, favorited) {
  var db = DBManager.db(recipient_id);
  var stmt = db.prepare("UPDATE statuses SET favorited = ? WHERE id = ?");
  stmt.bindInt32Parameter(0, favorited);
  stmt.bindStringParameter(1, status_id);
  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.logError("Failed to update favorite on local cache");
    db.rollback();
  }
};

Status.destroy = function(status_id, recipient_id) {
  var db = DBManager.db(recipient_id);
  var stmt0 = db.prepare("DELETE FROM statuses WHERE id = ?1");
  var stmt1 = db.prepare("DELETE FROM tweets   WHERE id = ?1");
  var stmt1 = db.prepare("DELETE FROM home     WHERE id = ?1");
  var stmt2 = db.prepare("DELETE FROM mentions WHERE id = ?1");
  stmt0.bindStringParameter(0, status_id);
  stmt1.bindStringParameter(0, status_id);
  stmt2.bindStringParameter(0, status_id);
  try {
    db.begin();
    stmt0.execute();
    stmt1.execute();
    stmt2.execute();
    db.commit();
  }
  catch (e) {
    db.log("Error on delete status");
    db.rollback();
  }
};

Status.getLatestId = function(recipient_id, type) {
  var db = DBManager.db(recipient_id);
  if (type != 'home' && type != 'mentions') return 0;

  var sql = "SELECT id FROM " + type + " ORDER BY id DESC LIMIT 1";
  var stmt = db.prepare(sql);

  var ret = 0;
  try {
    if (stmt.executeStep()) {
      return stmt.getString(0);
    }
  }
  catch (e) {
    db.logError("Failed to select " + type + " id");
  }
  return 0;
}

Status.blockUser = function(db_uid, user_id) {
  var db = DBManager.db(db_uid);

  var stmt = db.prepare("DELETE FROM statuses WHERE user_id = ?1");
  stmt.bindInt32Parameter(0, user_id);

  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.log("Error on delete status (block)");
    db.rollback();
  }
}


/////////////////////////////////////////////////////////////////////////
//
// Direct Message model
//
function DirectMessage(json, recipient_id)
{
  convertIdStr(json);
  for (var i in json) {
    this[i] = json[i];
  }
  this.user_id = recipient_id;
  this.db = DBManager.db(recipient_id);

  if (json) {
    this.sender = new User(this.sender, recipient_id);
    this.recipient = new User(this.recipient, recipient_id);
    this.user = this.sender;

    this.type = "messages";
    this.isSent = (this.sender.id == recipient_id);
  }
}

DirectMessage.prototype = {
  insertIntoDB: function() {
    var stmt = this.db.prepare("INSERT INTO direct_messages VALUES (?,?,?,?,?, ?)");

    stmt.bindStringParameter(0, this.id);
    stmt.bindStringParameter(1, this.sender_id);
    stmt.bindStringParameter(2, this.recipient_id);
    stmt.bindStringParameter(3, this.text);
    stmt.bindInt64Parameter(4, new Date(this.created_at).getTime());
    if (this.entities) {
      stmt.bindStringParameter(5, JSON.encode(this.entities));
    }

    var ret = true;
    try {
      stmt.execute();
    }
    catch (e) {
      if (this.db.lastError() != 19) {
        this.db.logError("Error on insert status");
        return false;
      }
      else {
        ret = false;
      }
    }

  if (0) {
    var fts = this.db.prepare("INSERT INTO messages VALUES (?, ?)");
    if (fts) {
      fts.bindStringParameter(0, this.id);
      fts.bindStringParameter(1, this.text);
      try {
        fts.execute();
      }
      catch (e) {
        EchofonUtils.debug("Failed to insert messages into FTS table");
      }
    }
  }

    this.sender.insertIntoDB();
    this.recipient.insertIntoDB();

    // Update threads table
    Thread.update(this);

    return ret;
  },

  toString: function() {
    return "DM: " + this.id + " " + this.sender.screen_name + " -> " + this.recipient.screen_name + " - " + this.text + " at " + new Date(this.created_at);
  }
};

DirectMessage.getConversationWith = function(db_user_id, recipient_id, count, offset)
{
  var results = [];

  if (!offset) offset = 0;
  if (!count) count = 20;
  if (recipient_id == db_user_id) {
    var sql = "SELECT * FROM direct_messages WHERE sender_id = ?1 AND recipient_id = ?1 ORDER BY id DESC LIMIT ?2 OFFSET ?3";
  }
  else {
    var sql = "SELECT * FROM direct_messages WHERE sender_id = ?1 OR recipient_id = ?1 ORDER BY id DESC LIMIT ?2 OFFSET ?3";
  }
  var stmt = DBManager.db(db_user_id).prepare(sql);

  stmt.bindInt64Parameter(0, recipient_id);
  stmt.bindInt32Parameter(1, count);
  stmt.bindInt32Parameter(2, offset);

  try {
    while (stmt.executeStep()) {
      var msg = new DirectMessage(null, db_user_id);
      // from statuses table
      msg.id                      = stmt.getString(0);
      msg.sender_id               = stmt.getString(1);
      msg.recipient_id            = stmt.getString(2);
      msg.text                    = stmt.getString(3);
      msg.created_at              = stmt.getInt64(4);
      try { msg.entities          = JSON.decode(stmt.getString(5))} catch (e) {};

      msg.sender    = User.findById(msg.sender_id, db_user_id);
      msg.recipient = User.findById(msg.recipient_id, db_user_id);

      msg.user = msg.sender;
      msg.type = "messages";

      results.push(msg);
    }
  }
  catch (e) {
    DBManager.db(db_user_id).logError("Failed to restore direct messages from local cache");
  }
  stmt.finalize();

  results.reverse();

  return results;
}

DirectMessage.destroy = function (message_id, db_user_id)
{
  var db = DBManager.db(db_user_id);
  var stmt = db.prepare("DELETE FROM direct_messages WHERE id = ?1");
  stmt.bindStringParameter(0, message_id);
  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.logError("Error on delete direct message");
    db.rollback();
  }
  stmt.finalize();
};

DirectMessage.getLatestId = function(recipient_id, type) {
  var stmt;
  var db = DBManager.db(recipient_id);
  if (type == "messages") {
      stmt = db.prepare("SELECT id FROM direct_messages WHERE recipient_id = ?1 AND sender_id != ?1 ORDER BY id DESC LIMIT 1");
      stmt.bindStringParameter(0, recipient_id);
  }
  else if (type == "sent") {
    stmt = db.prepare("SELECT id FROM direct_messages WHERE sender_id = ?1 AND recipient_id != ?1 ORDER BY id DESC LIMIT 1");
    stmt.bindStringParameter(0, recipient_id);
  }
  else {
    stmt = db.prepare("SELECT id FROM direct_messages ORDER BY id DESC LIMIT 1");
  }
  try {
    if (stmt.executeStep()) {
      return stmt.getString(0);
    }
  }
  catch (e) {
    db.logError("Failed to select " + type + " id");
  }
  return 0;
}

DirectMessage.getEarliestId = function(recipient_id, type) {
  var stmt;
  var db = DBManager.db(recipient_id);
  if (type == "messages") {
      stmt = db.prepare("SELECT id-1 FROM direct_messages WHERE recipient_id = ?1 AND sender_id != ?1 ORDER BY id LIMIT 1");
      stmt.bindStringParameter(0, recipient_id);
  }
  else if (type == "sent") {
    stmt = db.prepare("SELECT id-1 FROM direct_messages WHERE sender_id = ?1 AND recipient_id != ?1 ORDER BY id LIMIT 1");
    stmt.bindStringParameter(0, recipient_id);
  }
  try {
    if (stmt.executeStep()) {
      return stmt.getString(0);
    }
  }
  catch (e) {
    db.logError("Failed to select " + type + " id");
  }
  return 0;
}

DirectMessage.blockUser = function(db_uid, user_id) {
  var db = DBManager.db(db_uid);

  var stmt = db.prepare("DELETE FROM direct_messages WHERE sender_id = ?1 OR recipient_id = ?1");
  stmt.bindInt32Parameter(0, user_id);

  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.log("Error on delete direct message (block)");
    db.rollback();
  }
}

/////////////////////////////////////////////////////////////////////////
//
// Threads
//
function Thread(user_id)
{
  this.unread = 0;
  this.user_id = user_id;
}

Thread.prototype.toString = function()
{
  return "Thread: " + this.id + " - " + this.user.screen_name + " - " + this.text + "( " + this.created_at + ")";
}

Thread.initWithStatement = function(db_uid, stmt)
{
  var t = new Thread(db_uid);

  t.recipient_id  = stmt.getInt32(0);
  t.id            = stmt.getString(1);
  t.text          = stmt.getString(2);
  t.updated_at    = stmt.getInt64(3);

  t.type = "thread";

  t.user = User.findById(t.recipient_id, db_uid);
  return t;
}

Thread.update = function(message)
{
  var uid = 0;
  if (message.recipient_id == message.user_id) {
    uid = message.sender_id;
  }
  else {
    uid = message.recipient_id;
  }

  if(!Thread.checkNewMessage(uid, message)) {
    return;
  }

  var stmt = message.db.prepare("REPLACE INTO threads VALUES (?, ?, ?, ?)");
  stmt.bindInt32Parameter(0, uid);
  stmt.bindStringParameter(1, message.id);
  stmt.bindStringParameter(2, message.text);
  stmt.bindInt64Parameter(3, new Date(message.created_at).getTime());
  try {
    stmt.execute();
  }
  catch (e) {
    dump("Failed to replace thread: " + e.message + "\n");
  }
}

Thread.checkNewMessage = function(user_id, message)
{
  var stmt = message.db.prepare("SELECT id FROM threads WHERE recipient_id = ?");
  stmt.bindInt32Parameter(0, user_id);
  var prevId = 0;
  if (stmt.executeStep()) {
    prevId = stmt.getInt64(0);
  }
  return (prevId < message.id) ? true : false;
}

Thread.findByRecipientId = function(db_uid, recipient_id)
{
  var stmt = DBManager.db(db_uid).prepare("SELECT * FROM threads WHERE recipient_id = ?1");
  stmt.bindInt32Parameter(0, recipient_id);

  var t = null;

  try {
    if (stmt.executeStep()) {
      t = Thread.initWithStatement(db_uid, stmt);
    }
  }
  catch (e) {
    DBManager.db(db_uid).logError("Failed to get direct message threads from local cache");
  }
  return t;
}

Thread.restore = function(db_uid, count, max_id)
{
  var stmt;
  if (max_id) {
    var stmt = DBManager.db(db_uid).prepare("SELECT * FROM threads WHERE id <= ?2 ORDER BY id DESC LIMIT ?1");
    if (!count) count = 20;
    stmt.bindInt32Parameter(0, count);
    stmt.bindInt32Parameter(1, max_id);
  }
  else {
    var stmt = DBManager.db(db_uid).prepare("SELECT * FROM threads ORDER BY id DESC LIMIT ?1");
    if (!count) count = 20;
    stmt.bindInt32Parameter(0, count);
  }

  var results = [];

  try {
    while (stmt.executeStep()) {
      var t = Thread.initWithStatement(db_uid, stmt);
      results.push(t);
    }
  }
  catch (e) {
    DBManager.db(db_uid).logError("Failed to restore direct message threads from local cache");
  }
  return results;
}

Thread.blockUser = function(db_uid, user_id) {
  var db = DBManager.db(db_uid);

  var stmt = db.prepare("DELETE FROM threads WHERE recipient_id = ?1");
  stmt.bindInt32Parameter(0, user_id);

  try {
    db.begin();
    stmt.execute();
    db.commit();
  }
  catch (e) {
    db.log("Error on delete thread (block)");
    db.rollback();
  }
}

/////////////////////////////////////////////////////////////////////////
//
// Lists
//
function List(json, recipient_id)
{
  convertIdStr(json);
  for (var i in json) {
    this[i] = json[i];
  }
  this.recipient_id = recipient_id;
  this.db = DBManager.db(recipient_id);
  this.user = new User(this.user, this.recipient_id);
}

List.prototype = {
  insertIntoDB: function() {
    var stmt = this.db.prepare("REPLACE INTO lists VALUES (?,?,?,?,?, ?,?,?,?)");

    stmt.bindStringParameter(0, this.id);
    stmt.bindStringParameter(1, this.name);
    stmt.bindStringParameter(2, this.full_name);
    stmt.bindStringParameter(3, this.slug);
    stmt.bindStringParameter(4, this.description);
    stmt.bindInt32Parameter( 5, this.subscriber_count);
    stmt.bindInt32Parameter( 6, this.member_count);
    stmt.bindInt32Parameter( 7, this.mode);
    stmt.bindStringParameter(8, this.user.id);

    try {
      stmt.execute();
    }
    catch (e) {
      this.db.logError("Error on insert list");
      return false;
    }

    this.user.insertIntoDB();
    return true;
  },

  isOwnList: function() {
    return (this.user.id == this.recipient_id) ? true : false;
  }
};

List.findById = function(db_uid, list_id) {
  var stmt = DBManager.db(db_uid).prepare("SELECT * FROM lists WHERE id = ?");
  stmt.bindInt32Parameter(0, list_id);

  var list = null;
  try {
    while (stmt.executeStep()) {
      list = new List(null, db_uid);
      // from statuses table
      list.id                      = stmt.getString(0);
      list.name                    = stmt.getString(1);
      list.full_name               = stmt.getString(2);
      list.slug                    = stmt.getString(3);
      list.description             = stmt.getString(4);
      list.subscriber_count        = stmt.getInt32(5);
      list.member_count            = stmt.getInt32(6);
      list.mode                    = stmt.getInt32(7);
      list.user                    = User.findById(stmt.getString(8), db_uid);
    }
  }
  catch (e) {
    Components.utils.reportError("Failed to find list " + list_id + " from local cache");
  }
  return list;
}

List.destroy = function(list_id, db_uid) {
  var stmt = DBManager.db(db_uid).prepare("DELETE FROM lists WHERE id = ?");
  stmt.bindInt32Parameter(0, list_id);
  try {
    stmt.execute();
  }
  catch (e) {}
}

List.loadAll = function(recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("SELECT * FROM lists ORDER BY id DESC");

  var results = [];
  try {
    while (stmt.executeStep()) {
      var list = new List(null, recipient_id);
      // from statuses table
      list.id                      = stmt.getString(0);
      list.name                    = stmt.getString(1);
      list.full_name               = stmt.getString(2);
      list.slug                    = stmt.getString(3);
      list.description             = stmt.getString(4);
      list.subscriber_count        = stmt.getInt32(5);
      list.member_count            = stmt.getInt32(6);
      list.mode                    = stmt.getInt32(7);
      list.user                    = User.findById(stmt.getString(8), recipient_id);
      results.push(list);
    }
  }
  catch (e) {
    Components.utils.reportError("Failed to restore lists from local cache");
  }
  return results;
};

List.deleteAll = function(recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("DELETE FROM lists");
  try {
    stmt.execute();
  }
  catch (e) {
    DBManager.db(recipient_id).logError("Failed to delete lists from local cache");
  }
};


/////////////////////////////////////////////////////////////////////////
//
// Saved searches
//

function SavedSearch(json, recipient_id)
{
  convertIdStr(json);
  for (var i in json) {
    this[i] = json[i];
  }
  this.recipient_id = recipient_id;
  this.db = DBManager.db(recipient_id);
}

SavedSearch.prototype = {
  insertIntoDB: function() {
    var stmt = this.db.prepare("INSERT INTO saved_searches VALUES (?,?)");

    stmt.bindInt32Parameter(0, this.id);
    stmt.bindStringParameter(1, this.query);

    try {
      this.db.begin();
      stmt.execute();
      this.db.commit();
    }
    catch (e) {
      this.db.rollback();
      this.db.logError("Error on insert list");
      return false;
    }
    return true;
  }
};

SavedSearch.isExist = function(query, recipient_id)
{
  var stmt = DBManager.db(recipient_id).prepare("SELECT id FROM saved_searches WHERE query = ?");
  stmt.bindStringParameter(0, query);

  try {
    if (stmt.executeStep()) {
      return stmt.getInt32(0);
    }
  }
  catch (e) {}
  return 0;
}

SavedSearch.loadAll = function(recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("SELECT * FROM saved_searches ORDER BY id");

  var results = [];
  try {
    while (stmt.executeStep()) {
      var ss = new SavedSearch(null, recipient_id);
      // from statuses table
      ss.id                      = stmt.getInt32(0);
      ss.query                   = stmt.getString(1);
      results.push(ss);
    }
  }
  catch (e) {
    Components.utils.reportError("Failed to restore lists from local cache");
  }
  return results;
};

SavedSearch.deleteAll = function(recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("DELETE FROM saved_searches");
  try {
    stmt.execute();
  }
  catch (e) {
    DBManager.db(recipient_id).logError("Failed to delete all saved searches from local cache");
  }
};

SavedSearch.destroy = function(query_id, recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("DELETE FROM saved_searches WHERE id = ?");
  stmt.bindInt32Parameter(0, query_id);
  try {
    stmt.execute();
  }
  catch (e) {
    DBManager.db(recipient_id).logError("Failed to delete a saved search from local cache");
  }
};

/////////////////////////////////////////////////////////////////////////
//
// Block
//
function Blocking()
{
}

Blocking.deleteAll = function(recipient_id) {
  var stmt = DBManager.db(recipient_id).prepare("DELETE FROM blocks");
  try {
    stmt.execute();
  }
  catch (e) {}
}

Blocking.create = function(dbuid, user_id) {
  var stmt = DBManager.db(dbuid).prepare("INSERT INTO blocks VALUES (?)");
  stmt.bindInt32Parameter(0, user_id);
  try {
    stmt.executeStep();
  }
  catch (e) {}
}

Blocking.destroy = function(dbuid, user_id) {
  var stmt = DBManager.db(dbuid).prepare("DELETE FROM blocks WHERE id = ?");
  stmt.bindInt32Parameter(0, user_id);
  try {
    stmt.execute();
  }
  catch (e) {}
}

Blocking.isBlocking = function(dbuid, user_id) {
  var stmt = DBManager.db(dbuid).prepare("SELECT id FROM blocks WHERE id = ?");
  stmt.bindInt32Parameter(0, user_id);
  if (stmt.executeStep()) {
    return true;
  }
  else {
    return false;
  }
}


/////////////////////////////////////////////////////////////////////////
//
// No retweet
//
function NoRetweet()
{
}

NoRetweet.deleteAll = function(recipient_id) {
  try {
    var stmt = DBManager.db(recipient_id).prepare("DELETE FROM no_retweet");
    stmt.execute();
  }
  catch (e) {}
}

NoRetweet.create = function(dbuid, user_id) {
  try {
    var stmt = DBManager.db(dbuid).prepare("INSERT INTO no_retweet VALUES (?)");
    stmt.bindInt32Parameter(0, user_id);
    stmt.executeStep();
  }
  catch (e) {}
}

NoRetweet.destroy = function(dbuid, user_id) {
  try {
    var stmt = DBManager.db(dbuid).prepare("DELETE FROM no_retweet WHERE id = ?");
    stmt.bindInt32Parameter(0, user_id);
    stmt.execute();
  }
  catch (e) {}
}

NoRetweet.wantsRetweet = function(dbuid, user_id) {
  try {
    var stmt = DBManager.db(dbuid).prepare("SELECT id FROM no_retweet WHERE id = ?");
    stmt.bindInt32Parameter(0, user_id);
    if (stmt.executeStep()) {
      return false;
    }
    else {
      return true;
    }
  }
  catch (e) {
  }
  return true;
}
