//
// Implementation of Echofon timeline loader
//
// Copyright (c) 2010 Kazuho Okui / naan studio, Inc. ALL RIGHTS RESERVED.
//

var EXPORTED_SYMBOLS = ["TimelineLoader"];

const Cc = Components.classes;
const Ci = Components.interfaces;
const Cu = Components.utils;

const STATE_INIT                  = 0;  // no streaming API called yet
const STATE_STREAMING             = 1;  // using streaming API
const STATE_STREAMING_RETRYING    = 2;  // trying to reconnect to streaming server

Cu.import("resource://echofon/Models.jsm");
Cu.import("resource://echofon/EchofonUtils.jsm");
Cu.import("resource://echofon/EchofonSync.jsm");
Cu.import("resource://echofon/TwitterClient.jsm");

const HOME_TIMELINE     = 'home';
const MENTIONS_TIMELINE = 'mentions';
const DM_TIMELINE       = 'messages';
const LIST_TIMELINE     = 'lists';
const SEARCH_TIMELINE   = 'search';

var JSON = Cc["@mozilla.org/dom/json;1"].createInstance(Ci.nsIJSON);

//
// Timeline
//

function TimelineBase(user_id)
{
  this.user_id = user_id;
  this.tweets = [];
}

TimelineBase.prototype = {
  restore: function(count) {
    if (this.type) {
      this.tweets = Status.restore(this.user_id, this.type, count);
    }
  },

  insertTweet: function(tweet) {
    if (this.tweets.length == 0) {
      this.tweets.push(tweet);
      return true;
    }

    // check duplicate, insert with sort by id
    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.id == tweet.id) return false;
      if (parseInt(t.id) < parseInt(tweet.id)) {
	this.tweets.splice(0, 0, tweet);
        this.compaction();
	return true;
      }
    }
    this.tweets.push(tweet);
    this.compaction();
    return true;
  },

  sinceId: function() {
    for (var i in this.tweets) {
      if (this.tweets[i].user.id != this.user_id) return this.tweets[i].id;
    }
    return 0;
  },

  loadMore: function(count, max_id) {
    for (var i = 0; i < this.tweets.length; ++i) {
      if (DBManager._64bitsub(this.tweets[i].id, max_id) > 0) continue;
      return this.tweets.slice(i, i+count);
    }
    return null;
  },

  appendTweets: function(tweets) {
    for (var i in tweets) {
      this.tweets.push(tweets[i]);
    }
  },

  deleteTweet: function(tweetId) {
    var index = -1;
    for (var i =0; i < this.tweets.length; ++i) {
      var t = this.tweets[i];
      if (t.id == tweetId) {
        index = i;
        break;
      }
    }
    if (index >= 0) {
      this.tweets.splice(index, 1);
    }
  },

  toggleFavorite: function(tweetId, favorited) {
    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.id == tweetId) {
        t.favorited = favorited;
        break;
      }

      if (parseInt(t.id) < parseInt(tweetId)) {
        break;
      }
    }
  },

  blockUser: function(userId) {
    var statuses = [];

    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.user.id != userId) {
        statuses.push(t);
      }
    }
    delete this.tweets;
    this.tweets = statuses;
  },

  retweet: function(tweetId, tweet, is_undo) {
    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.id == tweetId) {
	if (is_undo) {
          t.retweeted_at = 0;
          t.retweeter_screen_name = "";
          t.retweeted_status_id = "";
          t.retweeter_user_id = 0;
	}
	else {
          t.retweeted_at = tweet.created_at;
          t.retweeter_screen_name = tweet.user.screen_name;
          t.retweeted_status_id = tweet.id;
          t.retweeter_user_id = tweet.user.id;
        }
        break;
      }

      if (parseInt(t.id) < parseInt(tweetId)) {
        break;
      }
    }
  },

  markRead: function() {
    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.unread == false) break;
      t.unread = false;
    }
  },

  syncUnread: function(id) {
    for (var i in this.tweets) {
      var t = this.tweets[i];
      if (t.unread == false) return;
      if (DBManager._64bitsub(id, t.id) >= 0) {
        t.unread = false;
      }
      else {
        return;
      }
    }
  },

  compaction: function() {
    if (this.tweets.length > 1000) {
      this.tweets.splice(500, 1000);
    }
  },

  toString: function() {
    return this.type + " timeline (" + this.tweets.length + "tweets)";
  }
};

//
// Home Timeline
//
function HomeTimeline(user_id)
{
  this.user_id = user_id;
  this.type = "home";
}

HomeTimeline.prototype = new TimelineBase();

//
// Mentions Timeline
//
function MentionsTimeline(user_id)
{
  this.user_id = user_id;
  this.type = "mentions";
  this.tweets = [];
}

MentionsTimeline.prototype = new TimelineBase();

//
// DM Timeline
//
function DMTimeline(user_id)
{
  this.user_id = user_id;
  this.type = "message";
  this.tweets = [];
}

DMTimeline.prototype = new TimelineBase();

DMTimeline.prototype.restore = function(count)
{
  this.tweets = Thread.restore(this.user_id, count);
}

DMTimeline.prototype.insertTweet = function(message)
{
  function sortThread(a, b) {
    return b.updated_at - a.updated_at;
  }
  var user = (message.isSent) ? message.recipient : message.sender;
  for (var i = 0; i < this.tweets.length; ++i) {
    var t = this.tweets[i];
    if (t.recipient_id == user.id) {
      var updated_at = new Date(message.created_at).getTime();
      if (t.updated_at < updated_at) {
	t.id = message.id;
        t.text = message.text;
        t.updated_at = updated_at;
      }
      if (message.unread) {
        t.unread += 1;
      }
      this.tweets.sort(sortThread);
      return true;
    }
  }

  var t = Thread.findByRecipientId(this.user_id, user.id);
  if (t) {
    if (message.unread) {
      t.unread +=1;
    }
    this.tweets.push(t);
    this.tweets.sort(sortThread);
  }
  return true;
}

DMTimeline.prototype.blockUser = function(userId)
{
  for (var i = 0; i < this.tweets.length; ++i) {
    var t = this.tweets[i];
    if (t.recipient_id == userId) {
      this.tweets.splice(i, 1);
      break;
    }
  }
}

DMTimeline.prototype.markRead = function(id)
{
  for (var i in this.tweets) {
    var t = this.tweets[i];
    if (t.id == id) {
      t.unread = false;
      return;
    }
  }
}



//
// List Timeline
//
function ListTimeline(user_id, list_id)
{
  this.user_id = user_id;
  this.list_id = list_id;
  this.type = "list";
  this.lists = {};
  this.tweets = [];
  this.changeList(list_id);
}

ListTimeline.prototype = new TimelineBase();

ListTimeline.prototype.changeList = function(list_id)
{
  // save current
  try {
    if (this.list_id) {
      this.lists[this.list_id] = this.tweets;
    }
  }
  catch (e) {}

  if (!list_id) {
    this.list_id = 0;
    return;
  }

  if (this.lists[list_id]) {
    this.tweets = this.lists[list_id];
  }
  else {
    this.lists[list_id] = [];
    this.tweets = [];
  }
  this.list_id = list_id;
}

ListTimeline.prototype.sinceId = function(list_id)
{
  if (!list_id || this.list_id == list_id) {
    return this.tweets.length ? this.tweets[0].id : 0;
  }
  else if (this.lists[list_id] && this.lists[list_id].length) {
    return this.lists[list_id][0].id;
  }
  return 0;
}

//
// Search Timeline
//
function SearchTimeline(user_id, query)
{
  this.user_id = user_id;
  this.query = query;
  this.type = "search";
  this.tweets = [];
}

SearchTimeline.prototype = new TimelineBase();


//
// Timelines
//

const gTimelineType = [HOME_TIMELINE, MENTIONS_TIMELINE, DM_TIMELINE, LIST_TIMELINE, SEARCH_TIMELINE];

function Timelines(token)
{
  this.token = token;
  var user_id = token.user_id;
  var list_id = 0;
  if (token.list_id) {
    list_id = token.list_id;
  }

  var query = '';
  if (token.query) {
    query = token.query;
  }

  this.timelines = {'home':new HomeTimeline(user_id),
		    'mentions':new MentionsTimeline(user_id),
		    'messages':new DMTimeline(user_id),
		    'lists':new ListTimeline(user_id, list_id),
		    'search':new SearchTimeline(user_id, query)};
}

Timelines.prototype = {
  tweets: function(type) {
    return this.timelines[type].tweets;
  },

  getRecentTweets: function(type) {
    var tweets = this.timelines[type].tweets;
    if (tweets.length == 0) return [];
    var ret = tweets.slice(0, 20);
    var last = ret[ret.length-1];
    if (last.unread == false) return ret;

    for (var i = ret.length; i < tweets.length; ++i) {
      var t = tweets[i];
      ret.push(tweets[i]);
      if (t.unread == false) {
        return ret;
      }
    }
    return ret;
  },

  getTimeline: function(type) {
    return this.timelines[type];
  },

  getUnread: function() {
    var ret = [];
    for (var i in this.timelines) {
      var tl = this.timelines[i];
      if (tl.tweets.length && tl.tweets[0].unread) {
        ret.push(true);
      }
      else {
        ret.push(false);
      }
    }
    return ret;
  },

  insert: function(tweet) {
    if (this.timelines[tweet.type]) {
      return this.timelines[tweet.type].insertTweet(tweet);
    }
    return false;
  },

  deleteTweet: function(tweetId) {
    this.timelines[HOME_TIMELINE].deleteTweet(tweetId);
    this.timelines[MENTIONS_TIMELINE].deleteTweet(tweetId);
  },

  blockUser: function(user_id) {
    for (var i in this.timelines) {
      this.timelines[i].blockUser(user_id);
    }
  },

  toggleFavorite: function(tweetId, favorited) {
    Status.toggleFavorite(this.token.user_id, tweetId, favorited);
    for (var i in this.timelines) {
      this.timelines[i].toggleFavorite(tweetId, favorited);
    }
  },

  retweet: function(tweetId, retweet, is_undo) {
    for (var i in this.timelines) {
      this.timelines[i].retweet(tweetId, retweet, is_undo);
    }
  },

  restore: function(tweet) {
    for (var i = 0; i < 3; ++i) {
      var type = gTimelineType[i];
      this.timelines[type].restore();
    }
  },

  updateSyncData: function(data) {
    if (data.user_id != this.token.user_id) return;
    var sync = data.data['sync'];

    var keys = {home:'timeline', mentions:'replies', messages:'messages'};
    for (var tab in keys) {
      var id = sync[keys[tab]];
      this.getTimeline(tab).syncUnread(id);
    }
    EchofonUtils.notifyObservers("updateDMThreads", {user_id:this.token.user_id, threads:this.tweets(DM_TIMELINE)});
  },

  compaction: function() {
    for (var i in this.timelines) {
      this.timelines[i].compaction();
    }
  }
};

//
// Timeline Loader
//

function TimelineLoader(token)
{
  this.token = token;
  this.db = DBManager.db(token.user_id);

  this.init();

  this.list = null;
  this.sentCount = 0;
  this.listLoaded = null;
  this.blocksLoaded = null;
  this.savedSearchesLoaded = null;
  this.query = token.query;
}

TimelineLoader.prototype = {

  init: function() {
    this.state = STATE_INIT;

    this.timelines = new Timelines(this.token);
    this.timelines.restore();

    this.streamingRetryPeriod = 0;
    this.terminated = false;
  },

  start: function() {

    if (this.state != STATE_INIT) return;

    if (!EchofonUtils.isActiveUser(this.token)) return;


    if (!this.getBlocking()) {
      this.startSession();
    }

    // Clear previous error message
    this.notifyObservers("clearErrors");
  },

  notifyObservers: function(command, params) {
    // do not send anything if this timeline is not active
    if (!EchofonUtils.isActiveUser(this.token)) return;
    EchofonUtils.notifyObservers(command, params);
  },

  startSession: function() {
    if (this.terminated) return;
    if (this.state == STATE_INIT) {
      this.state = STATE_STREAMING;
      this.userStream = this.startStreaming();

      this.refreshTimeline();
      this.refreshList();

      var target = this;
      var interval = (this.userStream) ? 10 : 1;

      this.timelineTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this.timelineTimer.initWithCallback({ notify: function() { target.refreshTimeline(); }}, interval * 60 * 1000, Ci.nsITimer.TYPE_REPEATING_SLACK);

      this.listTimer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this.listTimer.initWithCallback({ notify: function() { target.refreshList(); }}, 3 * 60 * 1000, Ci.nsITimer.TYPE_REPEATING_SLACK);
    }
  },

  refreshTimeline: function(force) {
    if (this.terminated) return;

    this.getTweets(HOME_TIMELINE);
    this.getTweets(MENTIONS_TIMELINE);
    this.getDMs("messages");

    var sent = true;
    if (!force && !this.userStream) {
      if (this.sentCount++ % 3 != 0) {
	sent = false;
      }
    }
    if (sent) {
      this.getDMs("sent");
    }
    this.getBlocking();

    if (!this.query) return;
    var now = Date.now();
    if (force || (!this.lastRefresh || now - this.lastRefresh > 3 * 60 * 1000)) {
      this.searchTweets(this.query, 0, true);
      this.lastRefresh = now;
    }
  },

  refreshList: function() {
    if (this.terminated) return;

    if (this.token.list_id) {
      var list = this.list;
      if (!list) {
	list = List.findById(this.token.user_id, this.token.list_id);
      }
      if (!list && !this.listLoaded) {
	  this.getLists();
      }
      else {
	this.getListTimeline(list);
      }
    }
  },

  getBlocking: function() {
    if (!this.userStream) return false;

    var now = Date.now();
    if (!this.blocksLoaded || now - this.blocksLoaded > 60 * 60 * 1000) {
      this.get('blocks.blocking.ids', {});
      this.get('friendships.no_retweet_ids', {});
      return true;
    }
    return false;
  },

  get: function(method, params, callback) {
    if (this.terminated) return;
    var req = new TwitterClient(this.token, this);
    req.get(method, params, callback);
  },

  getTweets: function(type, max_id) {
    if (this.terminated) return;

    var since_id = this.timelines.getTimeline(type).sinceId();
    if (since_id) {
      since_id = DBManager.calcSnowflakeId(since_id, 5);
    }
    var method = type == MENTIONS_TIMELINE ? 'mentions' : 'home_timeline';
    var params = {include_entities:true};
    if (max_id) {
        params['max_id'] = max_id;
    }
    else {
      if (since_id) {
        params['since_id'] = since_id;
        params['count'] = 200;
      }
    }
    this.get('statuses.' + method, params);
  },

  getDMs: function(type, max_id) {
    if (this.terminated) return;

    var since_id = DirectMessage.getLatestId(this.token.user_id, type);
    var method = type == 'messages' ? '' : '.sent';
    var params = {include_entities:true};
    if (max_id) {
      params['max_id'] = max_id;
    }
    else {
      if (since_id) {
        params['since_id'] = since_id;
        params['count'] = 200;
      }
    }
    this.get('direct_messages' + method, params);
  },

  getOlderDMs: function() {
    if (this.terminated) return;

    this.get('direct_messages',
             {include_entities:true, count:100,
              max_id:DirectMessage.getEarliestId(this.token.user_id, "messages")});

    this.get('direct_messages.sent',
             {include_entities:true, count:100,
              max_id:DirectMessage.getEarliestId(this.token.user_id, "sent")});

  },

  getLists: function() {
    if (this.terminated) return;

    var now = Date.now();
    if (!this.listLoaded || now - this.listLoaded > 60 * 60 * 1000) {
      this.get('lists.all', {});
    }
  },

  getSavedSearches: function() {
    if (this.terminated) return;

    var now = Date.now();
    if (!this.savedSearchesLoaded || now - this.savedSearchesLoaded > 60 * 60 * 1000) {
      this.get('saved_searches', {});
    }
  },

  getListTimeline: function(list, max_id) {
    if (this.terminated) return;

    if (list.id != this.token.list_id) return;

    var req = new TwitterClient(this.token, this);
    req.list = list;
    var params = {include_entities:true};
    if (max_id) {
        params['max_id'] = max_id;
    }
    else {
      this.list = list;
      var sinceId = this.timelines.getTimeline(LIST_TIMELINE).sinceId(list.id);
      if (sinceId) {
        params['since_id'] = sinceId;
        params['per_page'] = 200;
      }
    }
    params['user_id'] = list.user.id;
    params['list_id'] = list.id;
    this.get('lists.statuses', params);
  },

  lists_statuses: function(resp, req, context) {
    if (resp) {
      this.retrieveTweets(resp, LIST_TIMELINE, context);

      if (context.param.max_id && resp.length == 0) {
        this.notifyObservers("removeLoadMoreCell", 'mentions');
      }
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
 },

  isUnread: function(msg) {
    var sinceId = this.timelines.getTimeline(msg.type).sinceId();
    if (sinceId == 0) return false;

    if (EchofonSync.instance().isUnread(this.token.user_id, msg.type, msg.id)) {
      if (DBManager._64bitsub(sinceId, msg.id) >= 0) {
        return false;
      }
      else {
        return true;
      }
    }
    else {
      return false;
    }
  },

  retrieveTweets : function(obj, type, context) {
    if (obj == null) return;
    if (this.terminated) return;

    var tweets = [];

    var numUnread = 0;

    var db = DBManager.db(this.token.user_id);
    try {
      db.begin();
      for (var i in obj) {

        var msg;
        var result;

        if (type == HOME_TIMELINE && EchofonSync.instance().isMuted(this.token.user_id, obj[i])) continue;

        msg = new Status(obj[i], type, this.token.user_id);
        result = msg.insertIntoDB();
        // setup unread
        msg.unread = false;
        if (result) {
          if (this.isUnread(msg)) {
            msg.unread = true;
            numUnread++;
          }
        }
        tweets.push(msg);
      }
      db.commit();
    }
    catch (e) {
      Cu.reportError("Failed to insert " + this.type + " (" + e.message + ")");
      db.rollback();
    }
    EchofonUtils.debug("Received " + obj.length + " (" + numUnread + " new) " + type);

    var newTweets = [];
    for (var i = tweets.length-1; i >= 0; --i) {
      if (this.timelines.insert(tweets[i])) {
        newTweets.unshift(tweets[i]);
      }
    }
    if (context.param.max_id) {
      var ret = {type:type, msgs:tweets};
      this.notifyObservers("receivedNextPage", ret);
    }
    else {
      this.notifyObservers("receivedNewTweets", {user_id:this.token.user_id, tweets:newTweets, type:type});
    }
  },

  retrieveMessages : function(obj) {
    if (this.terminated) return;
    var messages = [];

    var db = DBManager.db(this.token.user_id);

    var numUnread = 0;

    try {
      db.begin();
      for (var i in obj) {

        var msg;
        var result;

        msg = new DirectMessage(obj[i], this.token.user_id);
        result = msg.insertIntoDB();
        if (!msg.isSent && this.isUnread(msg)) {
          msg.unread = true;
          numUnread++;
        }
	if (result) {
	  messages.push(msg);
	  this.timelines.insert(msg);
	}
      }
      db.commit();
    }
    catch (e) {
      Cu.reportError("Failed to insert direct messages (" + e.message + " " + e.fileName + " " + e.lineNumber + ")");
      db.rollback();
    }
    EchofonUtils.debug("Received " + obj.length + " (" + numUnread + " new) DM");

    this.notifyObservers("receivedNewTweets", {user_id:this.token.user_id, tweets:messages, type:DM_TIMELINE});
    this.notifyObservers("updateDMThreads", {user_id:this.token.user_id, threads:this.timelines.tweets(DM_TIMELINE)});
  },


  startStreaming: function() {
    this._timer = null;
    if (EchofonUtils.isActiveUser(this.token)) {
      try {
        this.stream = new TwitterClient(this.token, this);
        this.stream.stream(this.query);
        EchofonUtils.log("Start streaming" + ((this.query) ? " with query " + this.query : "") + " for " + this.token.screen_name + "...");
      }
      catch (e) {
        return false;
      }
    }
    return true;
  },

  stop: function() {
    this.terminated = true;

    if (this._timer) {
      this._timer.cancel();
      this._timer = null;
    }

    try {
      this.stream.abort();
    }
    catch (e) {}
    this.stream = null;
    if (this.timelineTimer) {
      this.timelineTimer.cancel();
      this.timelineTimer = null;
    }

    if (this.listTimer) {
      this.listTimer.cancel();
      this.listTimer = null;
    }

    this.timelines = null;
  },

  //
  // REST API callback
  //
  statuses_home_timeline : function(resp, request, context) {
    if (resp) {
      this.retrieveTweets(resp, HOME_TIMELINE, context);

      if (context.param.max_id && resp.length == 0) {
        this.notifyObservers("removeLoadMoreCell", 'home');
      }
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  statuses_mentions : function(resp, request, context) {
    if (resp) {
      this.retrieveTweets(resp, MENTIONS_TIMELINE, context);

      if (context.param.max_id && resp.length == 0) {
        this.notifyObservers("removeLoadMoreCell", 'mentions');
      }
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  direct_messages : function(resp, req, context) {
    if (resp) {
      this.retrieveMessages(resp);
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  direct_messages_sent : function(resp, req, context) {
    if (resp) {
      this.retrieveMessages(resp);
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  lists_all: function(resp, req, context) {
    if (resp) {
      List.deleteAll(this.token.user_id);
      for (var i in resp) {
        var list = new List(resp[i], this.token.user_id);
        list.insertIntoDB();
      }
      this.listLoaded = Date.now();
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  blocks_blocking_ids: function(resp, req, context) {
    if (resp) {
      Blocking.deleteAll(this.token.user_id);

      var db = DBManager.db(this.token.user_id);
      db.begin();
      for (var i in resp) {
        Blocking.create(this.token.user_id, resp[i]);
      }
      db.commit();

      this.blocksLoaded = Date.now();
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
    this.startSession();
  },

  friendships_no_retweet_ids: function(resp, req, context) {
    if (resp) {
      NoRetweet.deleteAll(this.token.user_id);
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
    var db = DBManager.db(this.token.user_id);
    db.begin();
    for (var i in resp) {
      NoRetweet.create(this.token.user_id, resp[i]);
    }
    db.commit();
  },

  saved_searches: function(resp, req, context) {
    if (resp) {
      SavedSearch.deleteAll(this.token.user_id);
      for (var i in resp) {
        var ss = new SavedSearch(resp[i], this.token.user_id);
        ss.insertIntoDB();
      }
      this.savedSearchesLoaded = Date.now();
      this.notifyObservers("updateSavedSearches", resp);
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  searchTweets: function(query, max_id, no_stream) {
    if (this.query != query) {
      this.timelines.getTimeline(SEARCH_TIMELINE).tweets = [];
    }
    this.query = query;
    var req = new TwitterClient(this.token, this);
    req.query = query;
    var params = {include_entities:true, q:query};
    if (max_id) {
        params['max_id'] = max_id;
    }
    else {
      var sinceId = this.timelines.getTimeline(SEARCH_TIMELINE).sinceId();
      if (sinceId) {
        params['since_id'] = sinceId;
        params['rpp'] = 100;
      }
    }
    req.endpoint = "search.twitter.com/";
    req.get("search", params);

    if (!max_id && !no_stream) {
      if (this._timer) {
        this._timer.cancel();
        this._timer = null;
      }

      if (this.searchStream) {
        this.searchStream.abort();
        this.searchStream = null;
      }

      this.searchStream = new TwitterClient(this.token, this);
      this.searchStream.stream(query);
    }
  },

  search: function(resp, req, context) {
    if (resp) {
      this.retrieveTweets(resp.results, SEARCH_TIMELINE, context);

      if (context.param.max_id && resp.results.length == 0) {
        this.notifyObservers("removeLoadMoreCell", SEARCH_TIMELINE);
      }

      if (resp.results.length == 0 && this.timelines.tweets(SEARCH_TIMELINE).length == 0) {
        this.notifyObservers("removeLoadMoreCell", SEARCH_TIMELINE);
      }
    }
    else {
      EchofonUtils.notifyObservers("APIError", context._errorMessage);
    }
  },

  clearSearch: function() {
    this.timelines.getTimeline(SEARCH_TIMELINE).tweets = [];
    this.query = null;
    this.startStreaming();
  },

  //
  // Streaming API callback
  //
  update_friends: function(friends, client) {
    EchofonUtils.notifyObservers("clearError");
    if (client == this.searchStream) {
      EchofonUtils.debug("============= Switch streaming connection =================");
      if (this.stream) {
        this.stream.abort();
      }
      this.stream = this.searchStream;
      this.searchStream = null;
    }
    this.streamingRetryPeriod = 0;
    this.state = STATE_STREAMING;

    EchofonUtils.debug("Received " + friends.length + " friends for " + this.token.screen_name);
    var delete_stmt = this.db.prepare("DELETE FROM following");
    delete_stmt.execute();

    this.db.begin();
    var insert_stmt = this.db.prepare("INSERT INTO following VALUES (?)");

    for (var i in friends) {
      var uid = friends[i];
      insert_stmt.bindInt32Parameter(0, friends[i]);
      insert_stmt.execute();
      insert_stmt.reset();
    }
    this.db.commit();
  },

  status_received: function(status) {
    var tweet = new Status(status, HOME_TIMELINE, this.token.user_id);

    // Do not display blocked user
    if (Blocking.isBlocking(this.token.user_id, tweet.user.id)) return;

    if (tweet.retweeter_user_id) {
      if (NoRetweet.wantsRetweet(this.token.user_id, tweet.retweeter_user_id) == false) {
        return;
      }
    }

    // If tweet isn't from following and doesn't have retweeted status, it should be mentions or search results
    //
    if (tweet.user.id != this.token.user_id && !tweet.user.isFollowing() && !tweet.retweeted_status_id) {
      if (tweet.has_mention) {
        tweet.type = MENTIONS_TIMELINE;
      }
      else {
        // if tweet has 'text@screen_name', it should be discarded
        var pat = new RegExp("@(" + this.token.screen_name + ")([^A-Za-z_].*)?$");
        if (pat.test(tweet.text)) {
          return;
        }
        tweet.type = SEARCH_TIMELINE;
      }
    }
    if (tweet.retweeter_user_id && !User.isFollowing(tweet.retweeter_user_id, this.token.user_id)) {
      tweet.type = SEARCH_TIMELINE;
    }

    // Avoid to insert search tweets after re-connectiing user streams
    if (tweet.type == SEARCH_TIMELINE && !this.query) return;

    tweet.unread = true;

    var isInHome = false;

    // insert status into home timeline only if user is not muted (mute is not applied to mentions and searches)
    if (tweet.type == HOME_TIMELINE && !EchofonSync.instance().isMuted(this.token.user_id, status)) {
      if (tweet.insertIntoDB()) {
        isInHome = true;
      }
    }
    if (isInHome || tweet.type != HOME_TIMELINE) {
      this.timelines.insert(tweet);
      this.notifyObservers("receivedNewTweets", {user_id:this.token.user_id, tweets:[tweet], type:tweet.type});
    }

    // Check mentions from following users.
    //
    if (tweet.type != MENTIONS_TIMELINE && tweet.has_mention) {
      var mention = new Status(status, MENTIONS_TIMELINE, this.token.user_id);
      mention.insertIntoDB();
      // do not mark read if it's already in home timeline
      if (isInHome == false) {
        mention.unread = true;
      }
      this.timelines.insert(mention);
      this.notifyObservers("receivedNewTweets", {user_id:this.token.user_id, tweets:[mention], type:MENTIONS_TIMELINE});
    }
  },

  direct_message_received: function(message) {
    this.retrieveMessages([message]);
  },

  event_received: function(obj) {
    switch (obj['event']) {
      case 'unfavorite':
      case 'favorite':
        this.timelines.toggleFavorite(obj['target_object'].id_str, (obj['event'] == 'favorite') ? true : false);
        this.notifyObservers("eventDidReceive", obj);
        break;

      case 'follow':
        if (obj.source.id == this.token.user_id) {
          User.follow(obj.target.id, this.token.user_id);
        }
        this.notifyObservers("eventDidReceive", obj);
        break;

      case 'block':
        // Needs to send block event to sync server
        Status.blockUser(this.token.user_id, obj.target.id);
        DirectMessage.blockUser(this.token.user_id, obj.target.id);
        Thread.blockUser(this.token.user_id, obj.target.id);
        this.timelines.blockUser(obj.target.id);
        Blocking.create(this.token.user_id, obj.target.id);
        this.notifyObservers("didBlockUser", obj.target);
        break;

      case 'unblock':
        // Send unblock event to sync server
        Blocking.destroy(this.token.user_id, obj.target.id);
        break;

      case 'list_created':
        var list = new List(obj.target_object, this.token.user_id);
        list.insertIntoDB();
        break;

      case 'list_destroyed':
        List.destroy(obj.target_object.id, this.token.user_id);
        break;

      case 'list_updated':
        var list = new List(obj.target_object, this.token.user_id);
        list.insertIntoDB();
        break;

      default:
        EchofonUtils.dumpall(obj);
        break;
    }
  },

  delete_received: function(obj) {
    if (obj['status']) {
      var info = obj['status'];
      Status.destroy(info.id_str, this.token.user_id);
      this.timelines.deleteTweet(info.id_str);
      this.notifyObservers("deleteTweet", {user_id:info.user_id_str, id:info.id_str});
    }
    else {
    }
  },

  retryStreaming: function(fillWithRESTAPI) {
    if (this.terminated) return;

    // retry user streams if appropriate
    if (this.state == STATE_STREAMING) {
      // if this is first failure, re-connect immediately
      if (fillWithRESTAPI) {
        this.refreshTimeline();
      }
      this.state = STATE_STREAMING_RETRYING;
      this.startStreaming();
      EchofonUtils.log("Re-connect User Streams immediately");
    }
    else if (this.state == STATE_STREAMING_RETRYING) {
      //wait a random period between 20 and 40 seconds.
      if (this.streamingRetryPeriod == 0) {
        this.streamingRetryPeriod = parseInt(Math.random() * 20 + 20);
      }
      else {
        this.streamingRetryPeriod *= 2;
        if (this.streamingRetryPeriod > 300) this.streamingRetryPeriod = 300;
      }
      EchofonUtils.log("Waiting for re-connct streaming server... " + this.streamingRetryPeriod);
      EchofonUtils.notifyObservers("APIError", "Re-connecting User Streams...");
      var target = this;
      this._timer = Cc["@mozilla.org/timer;1"].createInstance(Ci.nsITimer);
      this._timer.initWithCallback({
	notify: function() {
          // fill gap with REST API
          target.getTweets(HOME_TIMELINE);
          target["startStreaming"]()}
      }, this.streamingRetryPeriod * 1000, Ci.nsITimer.TYPE_ONE_SHOT);
    }
  },

  too_many_stream_connections_error: function() {
    this.notifyObservers("alertMessage", "Twitter has stopped allowing additional connections. This may happen if searches are made too frequently, or if you are using multiple Twitter apps at the same time.");
    this.state = STATE_STREAMING_RETRYING;
    this.streamingRetryPeriod = 60;
    this.retryStreaming(false);
  },

  stream_error: function(req) {
    if (this.stream == req || this.searchStream == req) {
      Cu.reportError("User stream connection is disconnected.");
      this.retryStreaming(false);
    }
  },

  stream_timeout: function() {
    this.retryStreaming(true);
  }
}
